#include <arpa/inet.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <inttypes.h>
#include <libgen.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>

#define _GNU_SOURCE
#include <string.h>

#include <unistd.h>

#include "buffer.h"
#include "sftp.h"
#include "utility.h"

#define BUFLEN 32768
#define SSH_PROGRAM "ssh"
#define STDIN 0
#define STDOUT 1
#define STDERR 2

static void     clear_attrs (ATTRS * attrs);
static void     destroy_ssh_commandline (char ** commandline);
static void     sftp_init (SftpConnection * conn, int in, int out);
static Handle * sftp_open (SftpConnection * conn, char * filename);
static char *   sftp_read (SftpConnection * conn, Handle * handle,
                           uint64_t offset, uint32_t flen, uint32_t * nread);
static ATTRS *  sftp_stat (SftpConnection * conn, char * filename);
static void     kill_child (int signo);
static char **  setup_ssh_commandline (char * hostname);
static int8_t   spawn_ssh (char * argv[], int * in, int * out);

static pid_t ssh_pid = -1;

int
main (int argc, char * argv[])
{
  ATTRS * attrs;
  Handle * handle;
  char * data;
  char * dstfile;
  char * hostname;
  char * srcfile;
  char ** ssh_argv;
  int in;
  int out;
  int outfile;
  int8_t status;
  uint32_t nread;
  uint32_t nwrite;
  uint64_t offset;
  SftpConnection conn;

  if (argc != 3)
    {
      fprintf (stderr, "%s: wrong arguments\n", argv[0]);
      return EXIT_FAILURE;
    }

  hostname = argv[1];
  srcfile = argv[2];
  dstfile = basename (srcfile);

  ssh_argv = setup_ssh_commandline (hostname);
  status = spawn_ssh (ssh_argv, &in, &out);
  if (status == -1)
    {
      fprintf (stderr, "failed to spawn\n");
      return EXIT_FAILURE;
    }

  sftp_init (&conn, in, out);
  attrs = sftp_stat (&conn, srcfile);
  if (attrs == NULL)
    {
      fprintf (stderr, "stat failed\n");
      return EXIT_FAILURE;
    }

  outfile = open (dstfile, O_CREAT | O_WRONLY | O_TRUNC,
                  S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
  if (outfile == -1)
    {
      fprintf (stderr, "failed to create destination file: %s\n", dstfile);
      return EXIT_FAILURE;
    }

  handle = sftp_open (&conn, srcfile);
  if (handle == NULL)
    {
      fprintf (stderr, "no handle\n");
      return EXIT_FAILURE;
    }

  offset = 0;
  data = sftp_read (&conn, handle, offset, BUFLEN, &nread);
  while (data != NULL)
    {
      nwrite = write (outfile, (void *) data, nread);
      if (nread != nwrite)
        fprintf (stderr, "nread(%u) != nwrite(%u)\n", nread, nwrite);

      free (data);
      offset += nread;
#ifdef DEBUG
      fprintf (stderr, "%llu bytes read\n", offset);
#endif
      data = sftp_read (&conn, handle, offset, BUFLEN, &nread);
    }

  close (outfile);

  free (data);
  free (handle->str);
  free (handle);
  free (attrs);
  destroy_ssh_commandline (ssh_argv);
  return EXIT_SUCCESS;
}

static void
clear_attrs (ATTRS * attrs)
{
  attrs->flags = htonl (0);
  attrs->size = 0;
  attrs->uid = 0;
  attrs->gid = 0;
  attrs->permissions = 0;
  attrs->atime = 0;
  attrs->mtime = 0;
  attrs->extended_count = 0;
}

static void
destroy_ssh_commandline (char ** commandline)
{
  uint32_t i;

  for (i = 0; i < 20 && commandline[i] != NULL; i++)
    free (commandline[i]);

  free (commandline);
}

static void
sftp_init (SftpConnection * conn, int in, int out)
{
  Buffer buf;
  char type;
  uint32_t plen;
  uint32_t version;

  conn->in = in;
  conn->out = out;
  conn->req_id = 0;

  plen = htonl (sizeof (type) + sizeof (version));
  type = SSH_FXP_INIT;
  version = htonl (SSH_FILEXFER_VERSION);

  buffer_init (&buf);

  buffer_add (&buf, (void *) &plen, sizeof (plen));
  buffer_add (&buf, (void *) &type, sizeof (type));
  buffer_add (&buf, (void *) &version, sizeof (version));

  buffer_write (&buf, conn->out);
  buffer_clear (&buf);

  buffer_read (&buf, conn->in);

  buffer_get (&buf, (void *) &plen, sizeof (plen));
  buffer_get (&buf, (void *) &type, sizeof (type));
  buffer_get (&buf, (void *) &version, sizeof (version));

  buffer_free (&buf);

  plen = ntohl (plen);
  version = ntohl (version);
}

static Handle *
sftp_open (SftpConnection * conn, char * filename)
{
  ATTRS attrs;
  Buffer buf;
  Handle * handle;
  char type;
  uint32_t plen;
  uint32_t slen;
  uint32_t pflags;
  uint32_t req_id;
  uint32_t status;

  handle = (Handle *) malloc (sizeof (Handle));
  if (handle == NULL)
    {
      fprintf (stderr, "SSH_FXP_OPEN: out of memory\n");
      return NULL;
    }

  plen = htonl (sizeof (type) + sizeof (req_id) + sizeof (slen)
         + strlen (filename) + sizeof (pflags) + sizeof (attrs.flags));

  type = SSH_FXP_OPEN;
  req_id = htonl (conn->req_id);
  slen = htonl (strlen (filename));
  pflags = htonl (SSH_FXF_READ);
  clear_attrs (&attrs);

  buffer_init (&buf);

  buffer_add (&buf, (void *) &plen, sizeof (plen));
  buffer_add (&buf, (void *) &type, sizeof (type));
  buffer_add (&buf, (void *) &req_id, sizeof (req_id));
  buffer_add (&buf, (void *) &slen, sizeof (slen));
  buffer_add (&buf, (void *) filename, strlen (filename));
  buffer_add (&buf, (void *) &pflags, sizeof (pflags));
  buffer_add (&buf, (void *) &(attrs.flags), sizeof (attrs.flags));

  buffer_write (&buf, conn->out);
  buffer_clear (&buf);

  buffer_read (&buf, conn->in);

  buffer_get (&buf, (void *) &plen, sizeof (plen));
  buffer_get (&buf, (void *) &type, sizeof (type));
  if (type == SSH_FXP_STATUS)
    {
      buffer_get (&buf, (void *) &status, sizeof (status));
      status = ntohl (status);
      fprintf (stderr, "SSH_FXP_OPEN: SSH_FXP_STATUS (%u)\n", status);
      return NULL;
    }

  buffer_get (&buf, (void *) &req_id, sizeof (req_id));
  req_id = ntohl (req_id);
  if (req_id != conn->req_id)
    {
      fprintf (stderr, "SSH_FXP_OPEN: wrong request id (%u) != %u\n",
               req_id, conn->req_id);
      return NULL;
    }

  buffer_get (&buf, (void *) &slen, sizeof (slen));
  handle->len = ntohl (slen);
  handle->str = (char *) malloc (handle->len * sizeof (char));
  if (handle->str == NULL)
    {
      fprintf (stderr, "SSH_FXP_OPEN: out of memory\n");
      return NULL;
    }

  buffer_get (&buf, (void *) handle->str, handle->len);

  buffer_free (&buf);

  plen = ntohl (plen);

  conn->req_id++;
  return handle;
}

static char *
sftp_read (SftpConnection * conn, Handle * handle, uint64_t offset,
           uint32_t flen, uint32_t * nread)
{
  Buffer buf;
  char * data = NULL;
  char type;
  uint32_t plen;
  uint32_t slen;
  uint32_t req_id;
  uint32_t status;

  plen = htonl (sizeof (type) + sizeof (req_id) + sizeof (slen)
         + handle->len + sizeof (offset) + sizeof (flen));

  type = SSH_FXP_READ;
  req_id = htonl (conn->req_id);
  slen = htonl (handle->len);
  offset = htonll (offset);
  flen = htonl (flen);

  buffer_init (&buf);

  buffer_add (&buf, (void *) &plen, sizeof (plen));
  buffer_add (&buf, (void *) &type, sizeof (type));
  buffer_add (&buf, (void *) &req_id, sizeof (req_id));
  buffer_add (&buf, (void *) &slen, sizeof (slen));
  buffer_add (&buf, (void *) handle->str, handle->len);
  buffer_add (&buf, (void *) &offset, sizeof (offset));
  buffer_add (&buf, (void *) &flen, sizeof (flen));

  buffer_write (&buf, conn->out);
  buffer_clear (&buf);

  buffer_read (&buf, conn->in);

  buffer_get (&buf, (void *) &plen, sizeof (plen));
  buffer_get (&buf, (void *) &type, sizeof (type));
  if (type == SSH_FXP_STATUS)
    {
      buffer_get (&buf, (void *) &status, sizeof (status));
      status = ntohl (status);
      fprintf (stderr, "SSH_FXP_READ: SSH_FXP_STATUS (%u)\n", status);
      return NULL;
    }

  buffer_get (&buf, (void *) &req_id, sizeof (req_id));
  req_id = ntohl (req_id);
  if (req_id != conn->req_id)
    {
      fprintf (stderr, "SSH_FXP_READ: wrong request id (%u) != %u\n",
               req_id, conn->req_id);
      return NULL;
    }

  buffer_get (&buf, (void *) &slen, sizeof (slen));
  flen = ntohl (flen);
  *nread = ntohl (slen);
  if (*nread == 0)
    {
      fprintf (stderr, "SSH_FXP_READ: EOF reached\n");
      return NULL;
    }
#ifdef DEBUG
  else if (*nread != flen)
    fprintf (stderr, "SSH_FXP_READ: nread(%u) != flen(%u)\n", *nread, flen);
#endif

  data = (char *) malloc (*nread * sizeof (char));
  if (data == NULL)
    {
      fprintf (stderr, "SSH_FXP_READ: out of memory\n");
      return NULL;
    }

  buffer_get (&buf, (void *) data, *nread);

  buffer_free (&buf);

  plen = ntohl (plen);

  conn->req_id++;
  return data;
}

static ATTRS *
sftp_stat (SftpConnection * conn, char * filename)
{
  ATTRS * attrs;
  Buffer buf;
  char type;
  uint32_t plen;
  uint32_t slen;
  uint32_t req_id;
  uint32_t status;

  attrs = (ATTRS *) malloc (sizeof (ATTRS));
  if (attrs == NULL)
    {
      fprintf (stderr, "SSH_FXP_STAT: out of memory\n");
      return NULL;
    }

  plen = htonl (sizeof (type) + sizeof (req_id) + sizeof (slen)
         + strlen (filename));

  type = SSH_FXP_STAT;
  req_id = htonl (conn->req_id);
  slen = htonl (strlen (filename));

  buffer_init (&buf);

  buffer_add (&buf, (void *) &plen, sizeof (plen));
  buffer_add (&buf, (void *) &type, sizeof (type));
  buffer_add (&buf, (void *) &req_id, sizeof (req_id));
  buffer_add (&buf, (void *) &slen, sizeof (slen));
  buffer_add (&buf, (void *) filename, strlen (filename));

  buffer_write (&buf, conn->out);
  buffer_clear (&buf);

  buffer_read (&buf, conn->in);

  buffer_get (&buf, (void *) &plen, sizeof (plen));
  buffer_get (&buf, (void *) &type, sizeof (type));
  if (type == SSH_FXP_STATUS)
    {
      buffer_get (&buf, (void *) &status, sizeof (status));
      status = ntohl (status);
      fprintf (stderr, "SSH_FXP_STAT: SSH_FXP_STATUS (%u)\n", status);
      return NULL;
    }

  buffer_get (&buf, (void *) &req_id, sizeof (req_id));
  req_id = ntohl (req_id);
  if (req_id != conn->req_id)
    {
      fprintf (stderr, "SSH_FXP_STAT: wrong request id (%u)\n", req_id);
      return NULL;
    }

  buffer_get (&buf, (void *) &(attrs->flags), sizeof (attrs->flags));
  attrs->flags = ntohl (attrs->flags);
  if (attrs->flags & SSH_FILEXFER_ATTR_SIZE)
    {
      buffer_get (&buf, (void *) &(attrs->size), sizeof (attrs->size));
      attrs->size = ntohll (attrs->size);
    }

  buffer_free (&buf);

  plen = ntohl (plen);

  conn->req_id++;
  return attrs;
}

static void
kill_child (int signo)
{
  if (ssh_pid > 0)
    {
      kill (ssh_pid, SIGTERM);
      waitpid (ssh_pid, NULL, 0);
    }
  _exit (EXIT_FAILURE);
}

static char **
setup_ssh_commandline (char * hostname)
{
  char ** argv;
  uint8_t index = 0;

  argv = (char **) malloc (sizeof (char *) * 20);
  if (argv == NULL)
    return NULL;

  argv[index++] = strdup (SSH_PROGRAM);
  argv[index++] = strdup ("-oForwardX11 no");
  argv[index++] = strdup ("-oForwardAgent no");
  argv[index++] = strdup ("-oClearAllForwardings yes");
  argv[index++] = strdup ("-oProtocol 2");
  argv[index++] = strdup ("-oNoHostAuthenticationForLocalhost yes");
  argv[index++] = strdup ("-s");
  argv[index++] = strdup (hostname);
  argv[index++] = strdup ("sftp");
  argv[index++] = NULL;

  return argv;
}

static int8_t
spawn_ssh (char * argv[], int * in, int * out)
{
  int pipein[2];
  int pipeout[2];
  int chin;
  int chout;

  if ((pipe (pipein) == -1) || (pipe (pipeout) == -1))
    return -1;

  *in = pipein[0];
  *out = pipeout[1];
  chin = pipeout[0];
  chout = pipein[1];  

  ssh_pid = fork();
  if (ssh_pid == -1)
    return -1;
  else if (ssh_pid == 0)
    {
      if ((dup2 (chin, STDIN) == -1) || (dup2 (chout, STDOUT) == -1))
        _exit (EXIT_FAILURE);

      close (*in);
      close (*out);
      close (chin);
      close (chout);

      signal (SIGINT, SIG_IGN);
      execvp (argv[0], argv);
      _exit (EXIT_FAILURE);
    }
  else
    {
      close (chin);
      close (chout);

      signal (SIGINT, kill_child);
      signal (SIGHUP, kill_child);
      signal (SIGTERM, kill_child);
    }

  return ssh_pid;
}
